#ifdef _WIN32
#include <windows.h>
#else
#include "../../WDL/swell/swell.h"
#endif

#include <stdio.h>
#include <math.h>

#include "resource.h"

#include "../reaper_plugin.h"

#include "../../WDL/wdlstring.h"
#include "../../WDL/jnetlib/jnetlib.h"

#define JNETLIB_WEBSERVER_WANT_UTILS
#include "../../WDL/jnetlib/webserver.h"

const char *(*get_ini_file)();
int (*GetNumTracks)();
int (*GetPlayState)();
double (*GetPlayPosition)();
double (*GetCursorPosition)();
int (*GetSetRepeat)(int val); // -1 == query, 0=clear, 1=set, >1=toggle . returns new value
void (*CSurf_OnRecord)();
void (*format_timestr)(double, char *, int);
const char *(*GetTrackInfo)(INT_PTR track, int *flags);
MediaTrack * (*CSurf_TrackFromID)(int, bool);
bool (*GetTrackUIVolPan)(MediaTrack *tr, double *vol, double *pan);
void *(*GetSetMediaTrackInfo)(MediaTrack *tr, const char *parmname, void *setNewValue);
void (*guidToString)(GUID *g, char *dest);
void (*Undo_OnStateChangeEx)(const char *descchange, int whichStates, int trackparm);
void (*CSurf_FlushUndo)(bool force);

double (*CSurf_OnVolumeChangeEx)(MediaTrack *trackid, double volume, bool relative, bool allowGang);
double (*CSurf_OnPanChangeEx)(MediaTrack *trackid, double pan, bool relative, bool allowGang);
bool (*CSurf_OnMuteChange)(MediaTrack *trackid, int mute);
bool (*CSurf_OnSelectedChange)(MediaTrack *trackid, int selected);
bool (*CSurf_OnSoloChange)(MediaTrack *trackid, int solo);
bool (*CSurf_OnFXChange)(MediaTrack *trackid, int en);
bool (*CSurf_OnRecArmChange)(MediaTrack *trackid, int recarm);



HWND g_main_hwnd;
REAPER_PLUGIN_HINSTANCE g_hInst;

char g_wwwroot[2048];
char g_config_userpass[1024];
int g_config_port=8808;
int g_config_cache_expire = 3600;


void typedResToString(void *res, const char *p, char *fmt, int fmt_size)
{
  fmt[0]=0;
  if (res) switch (*p)
  {
    case 'P':
      if (!strcmp(p,"P_NAME")) lstrcpyn(fmt,p,fmt_size);
    break;
    case 'B': sprintf(fmt,"%d",*(bool *)res ? 1 : 0); break; 
    case 'C': sprintf(fmt,"%d",*(char *)res); break; 
    case 'D': sprintf(fmt,"%f",*(double *)res); break; 
    case 'F': sprintf(fmt,"%f",*(float *)res); break; 
    case 'I': sprintf(fmt,"%d",*(int *)res); break; 
    case 'G':
      if (!strcmp(p,"GUID") || !strcmp(p,"G_GUID")) guidToString((GUID*)res,fmt);
    break;
  }
}

void *stringToTypedRes(const char *p, char *src, char *workbuf)
{
  switch (*p)
  {
    case 'P':
      if (!strcmp(p,"P_NAME")) return src;
    break;
    case 'B': 
      *(bool *)workbuf = !!atoi(src);
    return workbuf;  
    case 'C': 
      *(char *)workbuf = (char)atoi(src);
    return workbuf;  
    case 'F': 
      *(float *)workbuf = (float)atof(src);
    return workbuf;  
    case 'D': 
      *(double *)workbuf = (double)atof(src);
    return workbuf;  
    case 'I': 
      *(int *)workbuf = atoi(src);
    return workbuf;  
  }
  return NULL;
}

void ProcessCommand(WDL_String *o, const char *in, int inlen)
{
  char want[1024];
  lstrcpyn(want,in,min(sizeof(want),inlen+1));

  int cmd = atoi(want);
  if (cmd) 
  {
    if (cmd == 1013) CSurf_OnRecord();
    else SendMessage(g_main_hwnd,WM_COMMAND,cmd,0);
  }
  else if (!strncmp(want,"SET/",4))
  {
    if (!strncmp(want+4,"UNDO/",5))
    {
      CSurf_FlushUndo(true);
      Undo_OnStateChangeEx(want+9,UNDO_STATE_TRACKCFG|UNDO_STATE_MISCCFG,-1);
    }
    else if (!strncmp(want+4,"REPEAT/",7))
    {
      int a = atoi(want+11);
      if (a<0) a=2;
      else if (a) a=1;
      GetSetRepeat(a);
    }
    else if (!strncmp(want+4,"TRACK/",6))
    {
      char *p = want+10;
      MediaTrack *tr = CSurf_TrackFromID(atoi(p),false);

      if (tr)
      {
        while (*p && *p != '/') p++;
        while (*p == '/') p++;
        char *nm = p;
        while (*p && *p != '/') p++;
        if (*p)
        {
          *p++=0;
          while (*p == '/') p++;

          if (!strcmp(nm,"VOL")) CSurf_OnVolumeChangeEx(tr,atof(p),p[0]=='-'||p[0]=='+',p[strlen(p)-1]!='g');
          else if (!strcmp(nm,"PAN")) CSurf_OnPanChangeEx(tr,atof(p),p[0]=='-'||p[0]=='+',p[strlen(p)-1]!='g');
          else if (!strcmp(nm,"MUTE")) CSurf_OnMuteChange(tr,atoi(p));
          else if (!strcmp(nm,"SOLO")) CSurf_OnSoloChange(tr,atoi(p));
          else if (!strcmp(nm,"FX")) CSurf_OnFXChange(tr,atoi(p));
          else if (!strcmp(nm,"RECARM")) CSurf_OnRecArmChange(tr,atoi(p));
          else if (!strcmp(nm,"SEL")) CSurf_OnSelectedChange(tr,atoi(p));          
          else
          {
            char tmp[1024];
            void *setptr = stringToTypedRes(nm,p,tmp);
            if (setptr) GetSetMediaTrackInfo(tr,nm,setptr);
          }
        }
      }
    }
  }
  else if (!strncmp(want,"GET/",4))
  {
    if (!strncmp(want+4,"TRACK/",6))
    {
      const char *p = want+10;
      MediaTrack *tr = CSurf_TrackFromID(atoi(p),false);

      if (tr)
      {
        while (*p && *p != '/') p++;
        while (*p == '/') p++;
        void *res= *p ? GetSetMediaTrackInfo(tr,p,NULL) : NULL;
        
        o->Append(want);
        char fmt[1024];

        typedResToString(res,p,fmt,sizeof(fmt));

        o->Append(fmt);
        o->Append("\n");
      }
    }
    else if (!strcmp(want+4,"REPEAT"))
    {
      o->AppendFormatted(64,"GET/REPEAT\t%d\n",!!GetSetRepeat(-1));
    }
  }
  else if (!strcmp(want,"TRANSPORT"))
  {
    int ps = GetPlayState();
    double ppos=(ps&1) ? GetPlayPosition() : GetCursorPosition();
    //char tsbuf[512];
    //format_timestr(ppos,tsbuf,sizeof(tsbuf));
    o->AppendFormatted(1024,"TRANSPORT\t%d\t%f\t%d\n",ps,ppos,GetSetRepeat(-1));
  }
  else if (!strcmp(want,"NTRACK"))
  {
    o->AppendFormatted(64,"NTRACK\t%d\n",GetNumTracks());
  }
  else if (!strcmp(want,"TRACK") || !strncmp(want,"TRACK/",6))
  {
    int nt=GetNumTracks()+1;
    int minv=0;
    int maxv = nt;
    if (want[5]=='/')
    {
      char *p=want+6;
      minv = atoi(p);
      while (*p && *p != '-') p++;
      while (*p == '-') p++;
      maxv=(*p ? atoi(p) : minv)+1;
    }
    if (minv<0) minv=0;
    if (maxv>nt) maxv=nt;
    for (;minv<maxv;minv++)
    {
      int flag=0;
      MediaTrack *tr = CSurf_TrackFromID(minv,false);
      const char *nm=GetTrackInfo(minv-1,&flag);
      if (nm && tr)
      {
        char tname[512];
        lstrcpyn(tname,minv ? nm : "MASTER",sizeof(tname));
        char *p=tname;
        while (*p)
        {
          if (*p == '\t' || *p == '\n') *p=' ';
          p++;
        }
        double vol=1.0,pan=0.0;
        GetTrackUIVolPan(tr,&vol,&pan);

        
        int *peakhold = (int*)GetSetMediaTrackInfo(tr,"I_PEAKINFO_H",NULL);
        int *peakpos = (int*)GetSetMediaTrackInfo(tr,"I_PEAKINFO",NULL);

        o->AppendFormatted(256,"TRACK\t%d\t%s\t%d\t%f\t%f\t%d\t%d\n",minv,tname,flag,vol,pan,peakhold?*peakhold:-120000,peakpos?*peakpos:-120000);
      }
    }

  }

}



class wwwServer : public WebServerBaseClass
{
public:
  wwwServer() { tmpgen=NULL; } 
  ~wwwServer() { delete tmpgen; }
  JNL_StringPageGenerator *tmpgen;

  virtual IPageGenerator *onConnection(JNL_HTTPServ *serv, int port)
  {
//      char buf[512];
//      sprintf(buf,"got request at %.3f",GetTickCount()/1000.0);
//      OutputDebugString(buf);

    serv->set_reply_header("Server:reaper_www/0.0");

    if (g_config_userpass[0])
    {
      char *auth = serv->getheader("Authorization");
      if (auth)
      {
        if (!strnicmp(auth,"Basic",5))
        {
          auth += 5;
          while (*auth == ' ' || *auth == '\t') auth++;
          char decbuf[512];
          decbuf[0]=0;
          base64decode(auth,decbuf,sizeof(decbuf));
          if (strcmp(decbuf,g_config_userpass)) auth=NULL;
        }
        else auth=NULL;
      } 
      if (!auth)
      {
        serv->set_reply_header("WWW-Authenticate: Basic realm=\"reaper_www\"");
        serv->set_reply_string("HTTP/1.1 401 Unauthorized");
        serv->set_reply_size(0);
        serv->send_reply();
        return 0;
      }
    }   
    const char *req = serv->get_request_file();
    if (!strncmp(req,"/_/",3))
    {
      req+=3;
      if (!tmpgen) tmpgen = new JNL_StringPageGenerator;
      while (*req == ';') req++;
      while (*req)
      {
        const char *np = req;
        while (*np && *np != ';') np++;
      
        ProcessCommand(&tmpgen->str,req,np-req);

        req = np;
        while (*req == ';') req++;
      }

      JNL_StringPageGenerator *res=NULL;
      if (tmpgen->str.Get()[0])
      {
        res=tmpgen;
        tmpgen=0;
      }

      serv->set_reply_header("Content-Type: text/plain");
      serv->set_reply_header("Cache-Control: no-cache, must-revalidate"); // HTTP/1.1
      serv->set_reply_header("Expires: Sat, 26 Jul 1997 05:00:00 GMT"); // Date in the past
      serv->set_reply_string("HTTP/1.1 200 OK");
      serv->set_reply_size(res ? strlen(res->str.Get()) : 0);
      serv->send_reply();
      return res;
    }
    else if (req[0] == '/' && !strstr(req,".."))
    {
      if (!req[1]) req="/index.html";
      char buf[4096];
      lstrcpyn(buf,g_wwwroot,1024);
      lstrcpyn(buf+strlen(buf),req,1024);
      //OutputDebugString(buf);
      WDL_FileRead *fr = new WDL_FileRead(buf,0,32768,2);
      if (fr->IsOpen())
      {
        char ctstr[512];
        strcpy(ctstr,"Content-Type: ");
        JNL_get_mime_type_for_file(buf,ctstr+strlen(ctstr),256);
        serv->set_reply_header(ctstr);
        
        if (g_config_cache_expire)
        {
          char timefmt[512];
          strcpy(timefmt,"Expires: ");
          char *p = timefmt+strlen(timefmt);
          JNL_Format_RFC1123(time(NULL)+g_config_cache_expire,p);
          if (p[0]) serv->set_reply_header(timefmt);
        }
        else
        {
          serv->set_reply_header("Cache-Control: no-cache, must-revalidate"); // HTTP/1.1
          serv->set_reply_header("Expires: Sat, 26 Jul 1997 05:00:00 GMT"); // Date in the past
        }

        serv->set_reply_string("HTTP/1.1 200 OK");
        serv->set_reply_size((int)fr->GetSize());
        serv->send_reply();
        return new JNL_FilePageGenerator(fr);
      }
      else delete fr;
    }

    serv->set_reply_string("HTTP/1.1 404 NOT FOUND");
    serv->set_reply_size(0);
    serv->send_reply();
    return 0; // no data
  }
};

wwwServer *g_server;

UINT_PTR g_timer;

void CALLBACK TimerProc(
  HWND hwnd,     // handle of window for timer messages
  UINT uMsg,     // WM_TIMER message
  UINT_PTR idEvent,  // timer identifier
  DWORD dwTime   // current system time
)
{
  if (g_server) 
  {
    g_server->run();
  }
}


extern "C"
{

REAPER_PLUGIN_DLL_EXPORT int REAPER_PLUGIN_ENTRYPOINT(REAPER_PLUGIN_HINSTANCE hInstance, reaper_plugin_info_t *rec)
{
  g_hInst=hInstance;
  if (rec)
  {
    if (rec->caller_version != REAPER_PLUGIN_VERSION || !rec->GetFunc)
      return 0;

    HWND (*GetMainHwnd)();
#define IMPORT(x) if (!(*((void **)&x) = rec->GetFunc(#x))) return 0;

    IMPORT(CSurf_OnRecord)
    IMPORT(get_ini_file)
    IMPORT(guidToString)
    IMPORT(GetMainHwnd)
    IMPORT(GetPlayState)
    IMPORT(GetCursorPosition)
    IMPORT(GetSetRepeat)
    IMPORT(GetTrackInfo)
    IMPORT(CSurf_TrackFromID)
    IMPORT(GetTrackUIVolPan)
    IMPORT(GetSetMediaTrackInfo)
    IMPORT(GetPlayPosition)
    IMPORT(format_timestr)
    IMPORT(GetNumTracks)
    IMPORT(Undo_OnStateChangeEx)
    IMPORT(CSurf_FlushUndo)

   IMPORT(CSurf_OnVolumeChangeEx)
   IMPORT(CSurf_OnPanChangeEx)
   IMPORT(CSurf_OnMuteChange)
   IMPORT(CSurf_OnSelectedChange)
   IMPORT(CSurf_OnSoloChange)
   IMPORT(CSurf_OnFXChange)
   IMPORT(CSurf_OnRecArmChange)

#undef IMPORT

    GetPrivateProfileString("reaper_www","userpass","",g_config_userpass,sizeof(g_config_userpass),get_ini_file());
    g_config_port = GetPrivateProfileInt("reaper_www","port",g_config_port,get_ini_file());
    g_config_cache_expire = GetPrivateProfileInt("reaper_www","cacheexpire",g_config_cache_expire,get_ini_file());

    GetPrivateProfileString("reaper_www","wwwroot","",g_wwwroot,sizeof(g_wwwroot),get_ini_file());
    if (!g_wwwroot[0])
    {
      char *p=g_wwwroot;
      GetModuleFileName(hInstance,g_wwwroot,sizeof(g_wwwroot)-64);
      while (*p) p++;
      while (p >= g_wwwroot && *p != '\\' && *p != '/') p--;
      strcpy(++p,"reaper_www_root");
    }


    g_main_hwnd = GetMainHwnd();

    JNL::open_socketlib();

    g_server = new wwwServer;
    g_server->addListenPort(g_config_port);

    g_timer = SetTimer(NULL,0,20,(TIMERPROC)TimerProc);

    return 1;
  }
  else
  {
    KillTimer(NULL,g_timer);
    delete g_server;
    g_server=0;
    return 0;
  }
}

};

#ifndef _WIN32 // MAC resources
#include "../../WDL/swell/swell-dlggen.h"
#include "res.rc_mac_dlg"
#undef BEGIN
#undef END
#include "../../WDL/swell/swell-menugen.h"
#include "res.rc_mac_menu"
#endif